@@ -3,16 +3,18 @@ package HTML::Transmorgify::Metatags;
 
 use strict;
 use warnings;
-use HTML::Transmorgify qw(dangling eat_cr %variables $rbuf compile continue_compile capture_compile run $debug $modules queue_intercept rbuf boolean bomb $input_file $input_line $original_file);
+use HTML::Transmorgify qw(dangling eat_cr %variables $rbuf compile continue_compile capture_compile run $debug $modules queue_intercept rbuf boolean bomb $input_file $input_line $original_file allocate_result_type);
 use Scalar::Util qw(reftype blessed);
 use HTML::Entities;
 use URI::Escape;
 use List::Util;
 use File::Slurp;
+use Object::Dependency;
+use Carp qw(confess);
 require Exporter;
 
 our @ISA = qw(HTML::Transmorgify Exporter);
-our @result_array;
+our @metatags_result_array;
 our @include_dirs = ('.');
 our @EXPORT = ();
 our @EXPORT_OK = qw(%transformations @include_dirs %allowed_functions);
@@ -25,9 +27,10 @@ sub return_false { 0 };
 
 sub add_tags
 {
-	my ($self, $tobj) = @_;
-	$self->intercept_exclusive($tobj, __PACKAGE__, 100, %tags);
-	$self->intercept_shared($tobj, __PACKAGE__, 100, %shared_tags);
+	my ($module, $tobj) = @_;
+	$tobj->intercept_exclusive($tobj, __PACKAGE__, 100, %tags);
+	$tobj->intercept_shared($tobj, __PACKAGE__, 100, %shared_tags);
+	$tobj->intercept_pre_compile(\&setup_diversions);
 }
 
 $tags{"/define"} = \&dangling;
@@ -49,14 +52,19 @@ sub define_tag
 		bomb("illegal variable name for define") if $static_name =~ /^[\s_]/;
 		if (defined $static_value) {
 			# a common case, I expect
-			my $set = get_varset_func($static_name);
-			push(@$HTML::Transmorgify::rbuf, sub { $set->($static_value) });
+			push(@$HTML::Transmorgify::rbuf, sub { get_varset_func($static_name)->($static_value) });
 			eat_cr;
 			print STDERR "### DEFINE $static_name = '$static_value' DONE\n" if $debug;
 			return undef;
 		}
 	}
 
+
+	#
+	# Definitions can be in-line like <define name value />
+	# or they can be bracketed: <define name>value</define>.
+	# If they're in-line, then there will be a $raw_value.
+	#
 	my $buf;
 	if (defined $raw_value) {
 		$buf = compile($modules, \$raw_value);
@@ -77,7 +85,11 @@ sub define_tag
 		my $set = get_varset_func($name);
 
 		if ((@$buf == 1 && ! ref($buf->[0])) || ! $attr->boolean('eval', undef, 0)) {
-			# early binding -- let's set the variable to a string
+			#
+			# early binding -- let's set the variable to a string since the 
+			# the value is a simple string or we've been explicitly told
+			# not to evaluate the string.
+			#
 			my $vals = $attr->vals;
 			print STDERR "# attr $attr\n" if $debug;
 			$attr->hide_position(0) if $attr->last_position(0);
@@ -85,11 +97,20 @@ sub define_tag
 			{
 				printf STDERR "# local'izing variables{%s}\n", join(', ', keys %$vals) if $debug;
 				local(@HTML::Transmorgify::variables{keys %$vals}) = values %$vals;
-				local(@result_array) = ( '' );
-				run($buf, \@result_array);
-				$r = \@result_array;
+				local(@metatags_result_array) = ( '' );
+				run($buf, \@metatags_result_array);
+				$r = \@metatags_result_array;
 			}
 			if (@$r > 1) {
+				#
+				# Results can come in different "buckets".  Bucket 0, is the 
+				# main text results.  Other buckets are used for things like,
+				# HTTP headers, collecting javascript, etc.  Without knowing
+				# what they're used for it's hard to know what to do with 
+				# the extra results.  In any case, we're going to $set->() the
+				# new value to a callback.  XXX figure out semantics for 
+				# a macro expansion that adds to the HTTP header, etc.
+				#
 				$set->(sub {
 					$HTML::Transmorgify::result->[$_] .= $r->[$_] for grep { defined $r->[$_] } 0..$#$r;
 				});
@@ -100,6 +121,8 @@ sub define_tag
 			}
 		} else {
 			# late binding -- don't evaluate until the variable is expanded.
+			# we prefer early binding because more can happen at compile time and
+			# that will help performance if we can cache the results.
 			$set->(sub {
 				print STDERR "# evaluating variable '$name'...\n" if $debug;
 				my $vals = $attr->vals;
@@ -128,10 +151,8 @@ sub include_tag
 		print STDERR "# including file $attr\n" if $debug;
 		# $attr->run;
 		my $file = $attr->get('file', 0);
-		my $ifile = findfile($file);
+		my $ifile = find_file($file);
 		local($HTML::Transmorgify::input_file) = $ifile;
-#use Data::Dumper;
-#die Dumper($attr, $ifile, \@include_dirs) unless $ifile;
 		bomb("cannot find include file $file", attr => $attr) unless $ifile;
 		local($HTML::Transmorgify::input_line) = 1;
 		my $contents = read_file($ifile);
@@ -142,6 +163,144 @@ sub include_tag
 	return undef;
 };
 
+#
+# We're doing a little hack to grab control of the compile
+# sequence so that we can wrap the diversions code around it
+#
+
+our @process_text;
+
+sub setup_diversions
+{
+	push(@process_text, $HTML::Transmorgify::process_text_ref);
+	$HTML::Transmorgify::process_text_ref = \"<enableDiversionsCodeRightNow />";
+}
+
+$tags{enableDiversionsCodeRightNow} = \&enable_diversions;
+$tags{capture} = \&capture_tag;
+$tags{playback} = \&playback_tag;
+
+# because diversions can be nested, we have to 
+# resolve them in dependency order.  We'll use
+# an Object;:Dependency graph to store their names
+# so that they be resolved in dependency order.
+
+# At runtime, a playback tag cannot pull in the 
+# information it needs because it may not exist yet.
+# To solve that, the action of the playback tag is
+# to note the output possition within the current
+# block so that we can pull in the text in post-
+# processing.
+
+our $diversion_graph;
+our %diversion_text;
+our %diversion_used;
+our %diversion_playback;
+our $diversion_current;
+
+sub enable_diversions 
+{
+	my ($attr, $closed) = @_;
+	die unless $closed;
+	my $real_text = shift(@process_text);
+
+
+
+	rbuf (sub {
+		local(%diversion_text);
+		local(%diversion_used);
+		local(%diversion_playback);
+		local($diversion_graph) = Object::Dependency->new();
+
+		$diversion_graph->add('*main body*');
+
+		my $buf = compile($HTML::Transmorgify::modules, $real_text);
+		{
+			local($HTML::Transmorgify::result_array[0]) = '';
+			print STDERR "Start(diversion=*main body*)\n" if $debug;
+			local($diversion_current) = '*main body*';
+			run($buf);
+			$diversion_text{'*main body*'} = $HTML::Transmorgify::result_array[0];
+			print STDERR "End(diversion=*main body*)\n" if $debug;
+		}
+
+
+		#
+		# Pull the playbacks into the strings.
+		#
+		while(my ($ready) = $diversion_graph->independent(count => 1)) {
+			$diversion_graph->remove_dependency($ready);
+			my $textref = \$diversion_text{$ready};
+			$$textref = '' unless defined $$textref;
+			next unless $diversion_playback{$ready};
+			my @newstring;
+			my $lastpos = 0;
+			for my $playback (@{$diversion_playback{$ready}}) {
+				my $pos = $playback->{position};
+				push(@newstring, substr($$textref, $lastpos, $pos-$lastpos));
+				push(@newstring, $diversion_text{$playback->{diversion}});
+				$lastpos = $pos;
+			}
+			push(@newstring, substr($$textref, $lastpos, length($$textref)-$lastpos));
+			$$textref = join('', @newstring);
+		}
+
+		die unless $diversion_graph->alldone();
+
+		$HTML::Transmorgify::result_array[0] .= $diversion_text{'*main body*'};
+	});
+	return undef;
+}
+
+sub capture_tag 
+{
+	my ($attr, $closed) = @_;
+	die if $closed;
+	my $static_name = $attr->static('name', 0);
+	
+	my $diversion = \my $str;
+	my $buf = capture_compile("capture", $attr, $tag_package, '/capture' => \&return_false);
+	do_trim($attr, $buf);
+	eat_cr;
+
+	rbuf (sub {
+		my $name = $static_name or $attr->get('name', 0);
+		die "<capture> tag requires a capture name" unless defined $name;
+		allocate_result_type($name);
+		local($HTML::Transmorgify::result_array[0]) = '';
+		print STDERR "Start(diversion=$name)\n" if $debug;
+		local($diversion_current) = $name;
+		run($buf);
+		$diversion_text{$name} .= $HTML::Transmorgify::result_array[0];
+		print STDERR "End(diversion=$name)\n" if $debug;
+	});
+	return undef;
+}
+
+sub playback_tag 
+{
+	my ($attr, $closed) = @_;
+	warn "$attr is not closed, but it should be" unless $closed;
+	my $static_name = $attr->static('name', 0);
+
+	rbuf (sub {
+		my $name = $static_name or $attr->get('name', 0);
+		confess unless defined $diversion_current;
+		$diversion_playback{$diversion_current} ||= [];
+		push(@{$diversion_playback{$diversion_current}}, {
+			diversion	=> $name,
+			position	=> length($HTML::Transmorgify::result_array[0]),
+		});
+		$diversion_graph->add($diversion_current, $name);
+	});
+
+	return undef;
+}
+
+#
+# Add height & width to <img> tags if they don't 
+# already have them.
+#
 $shared_tags{img} = \&img_tag;
 sub img_tag 
 {
@@ -151,7 +310,8 @@ sub img_tag
 		print STDERR "# processing img $attr\n" if $debug;
 		my $vals = $attr->vals;
 		if (! $vals->{height} && ! $vals->{width} && $vals->{src}) {
-			$attr->set(attr_imgsize(findfile($attr->get('src'))) || die);
+			my $file = find_file($attr->get('src'));
+			$attr->set(attr_imgsize($file)) if $file;
 		}
 	});
 
@@ -159,29 +319,18 @@ sub img_tag
 }
 
 #
-# Inside <script> blocks, we care about "quoting".   In regular HTML,
-# we don't.
-#
-# XXXX NOT TRUE.   WE DO NOT CARE.  "</script>" will end the block
-# even if it is quoted.
+# Capture <script>...</script> and put it in the first result buffer
+# for inclusion in a combined <script> section.  The </script> will be
+# recognized even if it's inside a quoted string.
 #
 $tags{script} = \&script_tag;
-
 sub script_tag 
 {
 	my ($attr, $closed) = @_;
 	my $tag = "$attr";
 	my $before = pos($$HTML::Transmorgify::textref);
 	if (! $closed && $$HTML::Transmorgify::textref =~ m{
-		(?:
-			[^<'"]
-		|
-			' (?: [^\\'] | \.  )* '
-		|
-			" (?: [^\\"] | \.  )* "
-		|
-			< (?! /script [\s>] )
-		)*
+		(.*?)
 		</script [\s>]
 	}gcxs) {
 		my $text = substr($$HTML::Transmorgify::textref, $before, pos($$HTML::Transmorgify::textref)-$before);
@@ -205,6 +354,12 @@ our %transformations = (
 $tags{"/transform"} = \&dangling;
 $tags{transform} = \&transform_tag;
 
+#
+# XXX TODO items:
+#  1) cache transformation output
+#  2) notice when transforming constant inputs and do it at compile time.
+#
+
 sub transform_tag
 {
 	my ($attr, $closed) = @_;
@@ -232,6 +387,15 @@ sub transform_tag
 	});
 };
 
+#
+# If we set a value, like: <define foo.bar.3.baz>value</define>
+# then we have to to auto vivify an entire data structure.  We
+# assume that small (1-9999) integers are array indexes.
+#
+# We return a reference to the emtpy cell so that a value can 
+# be assigned.
+#
+
 sub create_container
 {
 	my ($ref, @refines) = @_;
@@ -245,6 +409,7 @@ sub create_container
 			$new = {};
 		}
 		if (blessed $ref) {
+			# should be a HTML::Transmorgify::ObjectGlue
 			$ref->set($ele, $new);
 		} elsif (ref($ref) eq 'ARRAY') {
 			$ref->[$ele] = $new;
@@ -258,6 +423,11 @@ sub create_container
 	return $ref;
 }
 
+#
+# This returns a function that sets a variable (of the $name
+# parameter) to a value.
+#
+
 sub get_varset_func
 {
 	my ($name) = @_;
@@ -265,6 +435,7 @@ sub get_varset_func
 		my ($cname, $ename) = ($1, $2);
 		my $container = lookup($cname, $ename, $cname);
 		if (blessed $container) {
+			# should be a HTML::Transmorgify::ObjectGlue
 			return sub { $container->set($ename, $_[0]) };
 		} elsif (ref($container) eq 'HASH') {
 			return sub { $container->{$ename} = $_[0] };
@@ -278,6 +449,12 @@ sub get_varset_func
 	}
 }
 
+#
+# This looks up a variable, optionally creating the container
+# object to to hold the variable.  Variables can have multiple
+# components to their names: eg: "a.b.c".  
+#
+
 sub lookup
 {
 	my ($name, $create, $error_context) = @_;
@@ -315,7 +492,6 @@ sub lookup
 		} elsif (ref $r) {
 			if (reftype($r) eq 'ARRAY') {
 				unless ($key =~ /\A\d+\z/) {
-					#return undef;
 					die "attempt to index into array with non-number '$key' for $error_context starting from $HTML::Transmorgify::original_file" 
 				}
 				$new = $r->[$key];
@@ -337,6 +513,11 @@ sub lookup
 	return $r;
 }
 
+#
+# to_text() returns the text associated with the lookup of a
+# variable.  
+#
+
 sub to_text
 {
 	my ($r, $attr) = @_;
@@ -367,32 +548,25 @@ sub to_text
 	return $r;
 }
 
-sub macro
-{
-	my ($attr) = @_;
 
-	my $name = $attr->get('name', 0);
-	my $encode = $attr->get('encode') || 'none';
-
-	printf STDERR "# macro '%s' encode='%s'\n", dstring($name), $encode if $debug;
-
-	my $r = to_text(lookup($name, 0, $attr), $attr);
-
-	die "no transformation '$encode' defined" unless $transformations{$encode};
-
-	printf STDERR "# before tranform %s: '%s'\n", $encode, $r if $debug;
-	$r = $transformations{$encode}->($r);
-	printf STDERR "# after  tranform %s: '%s'\n", $encode, $r if $debug;
-
-	return $r;
-}
 
 $tags{macro} = \&macro_tag;
-
 sub macro_tag {
 	my ($attr, $closed) = @_;
 	rbuf (sub {
-		my $r = macro($attr);
+		my $name = $attr->get('name', 0);
+		my $encode = $attr->get('encode') || 'none';
+
+		printf STDERR "# macro '%s' encode='%s'\n", dstring($name), $encode if $debug;
+
+		my $r = to_text(lookup($name, 0, $attr), $attr);
+
+		die "no transformation '$encode' defined" unless $transformations{$encode};
+
+		printf STDERR "# before tranform %s: '%s'\n", $encode, $r if $debug;
+		$r = $transformations{$encode}->($r);
+		printf STDERR "# after  tranform %s: '%s'\n", $encode, $r if $debug;
+
 		$HTML::Transmorgify::result->[0] .= $r if defined $r;
 		printf STDERR "# macro expansion '%s' = '%s'\n", "$attr", dstring($r) if $debug;
 	});
@@ -412,8 +586,8 @@ sub foreach_tag
 	my $buf = capture_compile("foreach", $attr, $tag_package, '/foreach' => \&return_false);
 	do_trim($attr, $buf);
 	eat_cr;
-# write test for same
-# document trim= for <foreach>
+	# write test for same
+	# document trim= for <foreach>
 
 	rbuf (sub {
 		my $var = $attr->get('var', 0);
@@ -436,6 +610,7 @@ sub foreach_tag
 			my $r = lookup($container, 0, $attr);
 			printf STDERR "# container value: %s.\n", dstring($r) if $debug;
 			if (blessed($r)) {
+				# should be a HTML::Transmorgify::ObjectGlue
 				my @e = $r->expand();
 				unless (@e == 1 && ref $r) {
 					my $i = 0;
@@ -579,7 +754,7 @@ sub conditional
 	}
 }
 
-sub findfile
+sub find_file
 {
 	my ($file) = @_;
 	die if $file =~ m{^/};
@@ -597,6 +772,13 @@ sub findfile
 }
 
 
+#
+# This is used by various control-flow tags to eliminate 
+# whitespace from around them.  You want to be able to format
+# the control-flow directives without adding whitespace to the
+# final document.  This function removes extra whitespace.
+#
+
 sub do_trim
 {
 	my ($attr, $buf) = @_;
